#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmshvert.h"

//-----------------------------------------------------------------------------------------
#define MSHVERT_EPSILON (0.00005f)
inline int mshvertSubtractFloats( float f1, float f2 )
{
   float f = f1 - f2;
   if( f < ( -MSHVERT_EPSILON ))
   {
      return -1;
   }
   else if( f > ( MSHVERT_EPSILON ))
   {
      return 1;
   }
   return 0;
}
#undef MSHVERT_EPSILON
//-----------------------------------------------------------------------------------------
CMshVert::CMshVert()
{
   mVertex = NULL;
   mNormal.x = 0.0f;
   mNormal.y = 0.0f;
   mNormal.z = 0.0f;
   mTx = 0;
   mTy = 0;
}
//-----------------------------------------------------------------------------------------
CMshVert::CMshVert( CMshVert *mv )
{
   mVertex = mv->mVertex;
   mNormal.x = mv->mNormal.x;
   mNormal.y = mv->mNormal.y;
   mNormal.z = mv->mNormal.z;
   mTx = mv->mTx;
   mTy = mv->mTy;
}
//-----------------------------------------------------------------------------------------
CMshVert::~CMshVert()
{
}
//-----------------------------------------------------------------------------------------
int CMshVert::Compare( CMshVert *arg1, CMshVert *arg2 )
{

   if( arg1->mVertex < arg2->mVertex )
   {
      return -1;
   }
   else if( arg1->mVertex > arg2->mVertex )
   {
      return 1;
   }
   else
   {
      int ret = mshvertSubtractFloats( arg1->mNormal.x, arg2->mNormal.x );
      if( ret == 0 )
      {
         ret = mshvertSubtractFloats( arg1->mNormal.y, arg2->mNormal.y );
         if( ret == 0 )
         {
            ret = mshvertSubtractFloats( arg1->mNormal.z, arg2->mNormal.z );
            if( ret == 0 )
            {
               ret = mshvertSubtractFloats( arg1->mTx, arg2->mTx );
               if( ret == 0 )
               {
                  ret = mshvertSubtractFloats( arg1->mTy, arg2->mTy );
               }
            }
         }
      }
      return ret;
   }

}
//-----------------------------------------------------------------------------------------
void mshvertQuickSort( CMshVert **items, int left, int right )
{

   int i, j;
   CMshVert *x, *y;
   i = left;
   j = right;
   x = items[( left + right ) / 2];

   do
   {
      while( CMshVert::Compare( items[i], x ) < 0 && i < right )
      {
         ++i;
      }
      while( CMshVert::Compare( x, items[j] ) < 0 && j > left )
      {
         --j;
      }
      if( i <= j )
      {
         y = items[i];
         items[i] = items[j];
         items[j] = y;
         ++i;
         --j;
      }
   } while( i <= j );

   if( left < j )
   {
      mshvertQuickSort( items, left, j );
   }
   if( i < right )
   {
      mshvertQuickSort( items, i, right );
   }

}
//-----------------------------------------------------------------------------------------
void CMshVert::Sort( CMshVert **items, int numitems )
{
   mshvertQuickSort( items, 0, numitems - 1 );
}
//-----------------------------------------------------------------------------------------
int CMshVert::Search( CMshVert **items, int numitems, CMshVert *key )
{

   int low, high, mid;
   low = 0;
   high = numitems - 1;

   while( low <= high )
   {
      mid = ( low + high ) / 2;
      int res = CMshVert::Compare( key, items[mid] );
      if( res < 0 )
      {
         high = mid - 1;
      }
      else if( res > 0 )
      {
         low = mid + 1;
      }
      else
      {
         return mid;
      }
   }

   return -1;

}
//-----------------------------------------------------------------------------------------
