#ifdef _WINDOWS
#ifndef WINDOWS
#define WINDOWS
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ac_plugin.h"
#include "cmshgroup.h"
#include "cmshmat.h"
#include "cmshtex.h"
#include "cmshvert.h"

//-----------------------------------------------------------------------------------------
void msh_getMatTex( ACObject *o, CMshMat **matarr, CMshTex **texarr, int &nummat, int &numtex )
{

   // Get the texture used, set its corresponding texarr tag to 1
   char tempstr[257];
   int idx;
   if( ac_object_has_texture( o ))
   {
      idx = ac_object_get_texture_index( o );
      if( idx > -1 )
      {
         if( texarr[idx]->mTag != 1 )
         {
            texarr[idx]->mTag = 1;
            texarr[idx]->SetName( texture_id_to_name( idx ));
#ifdef WINDOWS
            ac_unbackslash_string( texarr[idx]->mName );
#endif
            // Remove the file path
            char *pstr = strrchr( texarr[idx]->mName, '/' );
            if( pstr != NULL )
            {
               strcpy( tempstr, pstr + 1 );
               texarr[idx]->SetName( tempstr );
            }
         }
      }
   }

   // Get materials used, set their corresponding matarr tag to 1
   for( List *ps = ac_object_get_surfacelist( o ); ps != NULL; ps = ps->next )
   {
      idx = (( Surface * )( ps->data ))->col;
      if( idx > -1 )
      {
         if( matarr[idx]->mTag != 1 )
         {
            matarr[idx]->mTag = 1;
            ACEntity *m = ( ACEntity * )( ac_palette_get_material( idx ));
            ACrgb rgb;
            float f;
            char *str;
            ac_entity_get_rgb_value( m, "diffuse", &rgb );
            matarr[idx]->mDiffuse[0] = rgb.r;
            matarr[idx]->mDiffuse[1] = rgb.g;
            matarr[idx]->mDiffuse[2] = rgb.b;
            ac_entity_get_rgb_value( m, "ambient", &rgb );
            matarr[idx]->mAmbient[0] = rgb.r;
            matarr[idx]->mAmbient[1] = rgb.g;
            matarr[idx]->mAmbient[2] = rgb.b;
            ac_entity_get_rgb_value( m, "emissive", &rgb );
            matarr[idx]->mEmissive[0] = rgb.r;
            matarr[idx]->mEmissive[1] = rgb.g;
            matarr[idx]->mEmissive[2] = rgb.b;
            ac_entity_get_rgb_value( m, "specular", &rgb );
            matarr[idx]->mSpecular[0] = rgb.r;
            matarr[idx]->mSpecular[1] = rgb.g;
            matarr[idx]->mSpecular[2] = rgb.b;
            ac_entity_get_float_value( m, "shininess", &f );
            matarr[idx]->mSpecular[4] = f;
            ac_entity_get_float_value( m, "transparency", &f );
            matarr[idx]->mDiffuse[3] = min( 1.0f, max( 0.0f, 1.0f - f ));
            // TODO: Document that transparency applies to all material channels (using Diffuse for all transparency).
            matarr[idx]->mAmbient[3] = matarr[idx]->mDiffuse[3];
            matarr[idx]->mSpecular[3] = matarr[idx]->mDiffuse[3];
            matarr[idx]->mEmissive[3] = matarr[idx]->mDiffuse[3];
            ac_entity_get_string_value( m, "name", &str );
            matarr[idx]->SetName( str );
         }
      }
   }

   for( List *po = ac_object_get_childrenlist( o ); po != NULL; po = po->next )
   {
      msh_getMatTex(( ACObject * )( po->data ), matarr, texarr, nummat, numtex );
   }

}
//-----------------------------------------------------------------------------------------
void msh_addObjToMshGrp( ACObject *o, List **grparr, CMshMat **matarr, CMshTex **texarr )
{

   // Grab a count of verts
   int numverts = 0;
   for( List *ps = ac_object_get_surfacelist( o ); ps != NULL; ps = ps->next )
   {
      numverts += (( Surface * )( ps->data ))->numvert;
   }

   // List of all vertices + vertex normals + tex coords
   CMshVert **vertarr = new CMshVert*[numverts];
   CMshVert tmv;
   int ctr = 0;
   for( List *ps = ac_object_get_surfacelist( o ); ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      for( List *psv = s->vertlist; psv != NULL; psv = psv->next )
      {
         SVertex *sv = ( SVertex * )( psv->data );
         tmv.mVertex = sv->v;
         tmv.mNormal.x = sv->normal.x;
         tmv.mNormal.y = sv->normal.y;
         tmv.mNormal.z = sv->normal.z;
         tmv.mTx = sv->tx;
         tmv.mTy = sv->ty;
         vertarr[ctr++] = new CMshVert( &tmv );
      }
   }

   if( numverts > 0 )
   {

      // Sort the array
      CMshVert::Sort( vertarr, numverts );
      // Push only unique verts to a new array
      CMshVert *mvtemp = NULL;
      CMshVert **mshvert = new CMshVert*[numverts];
      int ctr = 0;
      for( int ix = 0; ix < numverts; ++ix )
      {
         if( ix == 0 )
         {
            mvtemp = vertarr[ix];
            mshvert[ctr++] = mvtemp;
         }
         else
         {
            if( CMshVert::Compare( mvtemp, vertarr[ix] ) == 0 )
            {
               // Same vertex - kick it out
               delete vertarr[ix];
            }
            else
            {
               mvtemp = vertarr[ix];
               mshvert[ctr++] = mvtemp;
            }
         }
      }
      delete[] vertarr;
      numverts = ctr;

      // Add geometry

      // Get triangle surfaces
      List *tris = ac_object_get_triangle_surfaces( o );
      int numtris = list_count( tris );

      // Cretae a group object
      CMshGroup *grp = new CMshGroup();

      // Grab the texture ID
      if( ac_object_has_texture( o ))
      {
         grp->mTexIdx = texarr[ac_object_get_texture_index( o )]->mTag;
      }

      // Allocate verts/tris
      grp->AllocateGeom( numverts, numtris );

      // Add verts
      for( int ix = 0; ix < numverts; ++ix )
      {
         grp->mVerts[ix][0] = mshvert[ix]->mVertex->x;
         grp->mVerts[ix][1] = mshvert[ix]->mVertex->y;
         grp->mVerts[ix][2] = mshvert[ix]->mVertex->z;
         grp->mVerts[ix][3] = mshvert[ix]->mNormal.x;
         grp->mVerts[ix][4] = mshvert[ix]->mNormal.y;
         grp->mVerts[ix][5] = mshvert[ix]->mNormal.z;
         grp->mVerts[ix][6] = mshvert[ix]->mTx;
         grp->mVerts[ix][7] = mshvert[ix]->mTy;
      }

      // Add tris
      int idx;
      ctr = 0;
      for( List *ps = tris; ps != NULL; ps = ps->next )
      {
         Surface *s = ( Surface * )( ps->data );
         // TODO: Document that we only get 1 material per object.  Maybe in a future release we can break up objects by material
         grp->mMatIdx = matarr[s->col]->mTag;
         idx = 0;
         for( List *psv = s->vertlist; psv != NULL; psv = psv->next )
         {
            SVertex *sv = ( SVertex * )( psv->data );
            tmv.mVertex = sv->v;
            tmv.mNormal.x = sv->normal.x;
            tmv.mNormal.y = sv->normal.y;
            tmv.mNormal.z = sv->normal.z;
            tmv.mTx = sv->tx;
            tmv.mTy = sv->ty;
            grp->mTris[ctr][idx] = CMshVert::Search( mshvert, numverts, &tmv );
            // Grab the material ID
            ++idx;
         }
         ++ctr;
      }

      // Cleanup
      list_free( &tris );
      for( int ix = 0; ix < numverts; ++ix )
      {
         delete mshvert[ix];
      }
      delete[] mshvert;

      // Add group entry
      list_add_item_head( grparr, grp );

   }

   for( List *po = ac_object_get_childrenlist( o ); po != NULL; po = po->next )
   {
      msh_addObjToMshGrp(( ACObject * )( po->data ), grparr, matarr, texarr );
   }

}
//-----------------------------------------------------------------------------------------
static Boolean do_msh_save( char *filename, ACObject *o )
{

   // Get material/texture lists
   CMshMat **matarr = NULL;
   CMshTex **texarr = NULL;
   int nummat = ac_get_num_palette();
   if( nummat > 0 )
   {
      matarr = new CMshMat*[nummat];
      // Add material tags
      for( int ix = 0; ix < nummat; ++ix )
      {
         matarr[ix] = new CMshMat();
      }
   }
   int numtex = ac_get_num_textures();
   if( numtex > 0 )
   {
      texarr = new CMshTex*[numtex];
      // Add texture tags
      for( int ix = 0; ix < numtex; ++ix )
      {
         texarr[ix] = new CMshTex();
      }
   }
   if( nummat > 0 || numtex > 0 )
   {
      msh_getMatTex( o, matarr, texarr, nummat, numtex );
   }

   // Renumber tags - non-zero numbers are valid entries
   int nummat_actual = 0;
   for( int ix = 0; ix < nummat; ++ix )
   {
      if( matarr[ix]->mTag == 1 )
      {
         matarr[ix]->mTag = ( ++nummat_actual );
      }
   }
   int numtex_actual = 0;
   for( int ix = 0; ix < numtex; ++ix )
   {
      if( texarr[ix]->mTag == 1 )
      {
         texarr[ix]->mTag = ( ++numtex_actual );
      }
   }

   // Push objects to group definitions
   List *grparr = NULL;
   msh_addObjToMshGrp( o, &grparr, matarr, texarr );

   FILE *fp = fopen( filename, "w" );
   if( fp == NULL) { 
      message_dialog( "error opening file \"%s\"", filename );
      return NULL;
   }

   Boolean success = FALSE;

   // Write the header
   fprintf( fp, "MSHX1\n" );
   fprintf( fp, "GROUPS %d\n", list_count( grparr ));

   // TODO: Sort the group array by texture, then by material.  Transparent textures/materials should be at the end of the mesh
   //       This will require making the group an array instead of a List

   // Write geometry
   for( List *p = grparr; p != NULL; p = p->next )
   {
      CMshGroup *grp = ( CMshGroup * )( p->data );
      fprintf( fp, "MATERIAL %d\n", grp->mMatIdx );
      if( grp->mTexIdx > 0 )
      {
         fprintf( fp, "TEXTURE %d\n", grp->mTexIdx );
      }
      else
      {
         fprintf( fp, "TEXTURE 0\n" );
      }
      fprintf( fp, "GEOM %d %d\n", grp->mNumVerts, grp->mNumTris );
      for( int iy = 0; iy < grp->mNumVerts; ++iy )
      {
         if( grp->mTexIdx > 0 )
         {
            fprintf( fp, "%g %g %g %g %g %g %g %g\n",
                  -( grp->mVerts[iy][0] ), grp->mVerts[iy][1], grp->mVerts[iy][2],
                  -( grp->mVerts[iy][3] ), grp->mVerts[iy][4], grp->mVerts[iy][5],
                   ( grp->mVerts[iy][6] ), -( grp->mVerts[iy][7] ));
         }
         else
         {
            fprintf( fp, "%g %g %g %g %g %g\n",
                  -( grp->mVerts[iy][0] ), grp->mVerts[iy][1], grp->mVerts[iy][2],
                  -( grp->mVerts[iy][3] ), grp->mVerts[iy][4], grp->mVerts[iy][5] );
         }
      }
      for( int iy = 0; iy < grp->mNumTris; ++iy )
      {
         fprintf( fp, "%d %d %d\n", grp->mTris[iy][2], grp->mTris[iy][1], grp->mTris[iy][0] );
      }
   }

   // Write materials
   fprintf( fp, "MATERIALS %d\n", nummat_actual );
   for( int ix = 0; ix < nummat; ++ix )
   {
      if( matarr[ix]->mTag > 0 )
      {
         // TODO: Ensure that we don't need the texture spec here, i.e., "Material3 with Tex: shutex" instead of just "Material3"
         fprintf( fp, "%s\n", matarr[ix]->mName );
      }
   }
   for( int ix = 0; ix < nummat; ++ix )
   {
      if( matarr[ix]->mTag > 0 )
      {
         // TODO: Ensure that we don't need the texture spec here, i.e., "Material3 with Tex: shutex" instead of just "Material3"
         fprintf( fp, "MATERIAL %s\n", matarr[ix]->mName );
         fprintf( fp, "%g %g %g %g\n", matarr[ix]->mDiffuse[0], matarr[ix]->mDiffuse[1], matarr[ix]->mDiffuse[2], matarr[ix]->mDiffuse[3] );
         fprintf( fp, "%g %g %g %g\n", matarr[ix]->mAmbient[0], matarr[ix]->mAmbient[1], matarr[ix]->mAmbient[2], matarr[ix]->mAmbient[3] );
         fprintf( fp, "%g %g %g %g %g\n", matarr[ix]->mSpecular[0], matarr[ix]->mSpecular[1], matarr[ix]->mSpecular[2], matarr[ix]->mSpecular[3], matarr[ix]->mSpecular[4] );
         fprintf( fp, "%g %g %g %g\n", matarr[ix]->mEmissive[0], matarr[ix]->mEmissive[1], matarr[ix]->mEmissive[2], matarr[ix]->mEmissive[3] );
      }
   }

   // Write textures
   fprintf( fp, "TEXTURES %d\n", numtex_actual );
   for( int ix = 0; ix < numtex; ++ix )
   {
      if( texarr[ix]->mTag > 0 )
      {
         fprintf( fp, "%s\n", texarr[ix]->mName );
      }
   }

   fclose( fp );

   // Cleanup
   for( int ix = 0; ix < nummat; ++ix )
   {
      delete matarr[ix];
   }
   for( int ix = 0; ix < numtex; ++ix )
   {
      delete texarr[ix];
   }
   for( List *p = grparr; p != NULL; p = p->next )
   {
      delete (( CMshGroup * )( p->data ));
   }
   if( matarr != NULL )
   {
      delete[] matarr;
   }
   if( texarr != NULL )
   {
      delete[] texarr;
   }
   list_free( &grparr );

   return success;

}
//-----------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *data )
{

   static bool firstTime = true;
   if( firstTime )
   {

      double ver = ac_get_version_number();

      firstTime = false;

      if (ver < 3.7)
      {
         printf( "ERROR - msh importer requires AC3D 3.7 or later" );
         return -1;
      }

      ac_register_file_exporter( "MSH", ".msh", "Orbiter MSH files", do_msh_save, "plugin, version 1.0, by Dennis Hawthorne" );

   }

   return 0;

}
//-----------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//-----------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Orbiter MSH Export - Version 1.0 - by Dennis Hawthorne" );
}
//-----------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Orbiter MSH exporter." );
}
//-----------------------------------------------------------------------------------------
